
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: _G, coroutine, Grid2Frame, Grid2Layout, Grid2LayoutFrame, IsAddOnLoaded, next

local InCombatLockdown = InCombatLockdown
local Mixin = Mixin

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Integration Grid2")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.numActiveButtons = 0

function M:OnEnable()
	if not IsAddOnLoaded("Grid2") then
		self:RegisterEvent("ADDON_LOADED")
		return
	end

	if not self:IsHooked("CreateFrame") then
		self:SecureHook("CreateFrame")
	end

	self:RegisterMessage("SIXUI_CONFIG_UPDATE")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED", "Update")
	self:RegisterMessage("SIXUI_PROFILE_UPDATE", "Update")
	self:HookAllButtons()
	self:Update()
end

function M:OnDisable()
	self:UnhookAllButtons()
end

function M:Update()
	self:AdjustButtons()
	self:AdjustLayoutFrame()
end

-- Determine the most relevant raid frame config.
function M:GetRaidFrameConfig()
	return A:GetRaidFrameConfig(self.numActiveButtons)
end

function M:AdjustButtons(tick)
	if self.timer then
		self:CancelTimer(self.timer)
	end

	if not tick then
		self.timer = self:ScheduleTimer("AdjustButtons", 0.1, true)
		return
	end

	if InCombatLockdown() then
		A.CombatQueue:Add(self.AdjustButtons, self.AdjustButtons, self, true)
	else
		local raidFrameConfig = self:GetRaidFrameConfig()
		local profile = Grid2Frame.db.profile

		Mixin(profile, {
			frameWidth = raidFrameConfig.width,
			frameHeight = raidFrameConfig.height,
		})

		Grid2Layout:UpdateFramesSize()
	end
end

function M:AdjustLayoutFrame()
	if InCombatLockdown() then
		A.CombatQueue:Add(self.AdjustLayoutFrame, self.AdjustLayoutFrame, self)
	else
		local raidFrameConfig = self:GetRaidFrameConfig()

		A:RegisterFrame("Grid2LayoutFrame"):ClearAllPoints():SetPoint(raidFrameConfig.point, raidFrameConfig.relativeTo, raidFrameConfig.relativePoint, raidFrameConfig.offsetX, raidFrameConfig.offsetY)
	end
end

function M:DecrementNumActiveButtons()
	self.numActiveButtons = self.numActiveButtons - 1
	self:AdjustButtons()
end

function M:IncrementNumActiveButtons()
	self.numActiveButtons = self.numActiveButtons + 1
	self:AdjustButtons()
end

-- Itereate over Grid2's buttons.
function M:IterateButtons()
	if (not Grid2LayoutFrame) then return end

	return coroutine.wrap(function ()
		for _, header in next, {Grid2LayoutFrame:GetChildren()} do
			if (header.GetName and header:GetName() or ""):match("Grid2LayoutHeader%d+") then
				for _, button in next, {header:GetChildren()} do
					if (button.GetName and button:GetName() or ""):match("Grid2LayoutHeader%d+UnitButton%d+") then
						coroutine.yield(button)
					end
				end
			end
		end
	end)
end

function M:HookAllButtons()
	for button in self:IterateButtons() do
		self:HookButton(button)
	end
end

function M:HookButton(button)
	if (not button) then return end

	if not self:IsHooked(button, "OnShow") then
		self:HookScript(button, "OnShow", "SkinButton")
	end

	if not self:IsHooked(button, "OnHide") then
		self:HookScript(button, "OnHide", "UnskinButton")
	end

	self[button:IsShown() and "SkinButton" or "UnskinButton"](self, button)
end

function M:UnhookAllButtons()
	for button in self:IterateButtons() do
		self:UnhookButton(button)
	end
end

function M:UnhookButton(button)
	self:Unhook(button, "OnHide")
	self:Unhook(button, "OnShow")
	self:UnskinButton(button)
end

function M:SkinButton(button)
	B:CreateBorder(button, C.BORDER_PADDING):SetIgnoreParentAlpha(true)

	if not self:IsHooked(button, "SetBackdropBorderColor") then
		self:RawHook(button, "SetBackdropBorderColor", "OnSetBackdropBorderColor", true)
		self.hooks[button].SetBackdropBorderColor(button, 0, 0, 0, 0)
		self:OnSetBackdropBorderColor(button, button:GetBackdropBorderColor())
		self:IncrementNumActiveButtons()
	end
end

function M:UnskinButton(button)
	B:RemoveBorder(button)

	if self:IsHooked(button, "SetBackdropBorderColor") then
		self:Unhook(button, "SetBackdropBorderColor")
		self:DecrementNumActiveButtons()
	end
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:ADDON_LOADED(event, name)
	if name == "Grid2" then
		self:UnregisterEvent("ADDON_LOADED")
		self:OnEnable()
	end
end

function M:SIXUI_CONFIG_UPDATE(event, module, config, key, ...)
	if module == "Core RaidFrameConfig" then
		if key == "width" or key == "height" then
			self:AdjustButtons()
		else
			self:AdjustLayoutFrame()
		end
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M:CreateFrame(frameType, frameName, parentFrame, inheritsFrame)
	if frameName then
		if frameName == "Grid2LayoutFrame" then
			self:AdjustLayoutFrame()
		elseif frameName:match("Grid2LayoutHeader%d+UnitButton%d+") then
			self:HookButton(_G[frameName])
		end
	end
end

function M:OnSetBackdropBorderColor(button, r, g, b)
	if button.GetBorder then
		if r ~= 0 or g ~= 0 or b ~= 0 then
			button:GetBorder():SetColor(r, g, b)
		else
			button:GetBorder():SetColor()
		end
	end
end
